;; N-body calculation using Greenspan 1st-order discrete mechanics.
;; 1-D version with vector operations, combines
;; particle and pair domains.  Particle calculations performed at
;; left edge instead of diagonal
;; Spreading and accumultion use V5.0 features.  Uses grid customized
;; for problem size

;; Provides for insertion of asteroids uniformly spaced in circular
;; orbit of specified radius, control over printing.

(*proclaim '(type fixnum N LN CN NB NP PX NX NA))

(defvar N  64)  ; Max number of bodies - power of 2, N/2 < NB <= N
(defvar LN  6)  ; log2(N)
(defvar CN 12)  ; Field size for coordinates
(defvar NB  0)  ; Actual number of bodies
(defvar NP  1)  ; Number of NxN blocks in VP array
(defvar NX  0)  ; Number of bodies given explicitly
(defvar NA  0)  ; Number of asteroids
(defvar PX  2)  ; Print control

;; Geometry for VP set, calculated by setup

(defvar GEOM nil)

(*proclaim '(type single-float G AR AS TWOPI hdt))

(defvar G  0.0)             ; Gravitational constant
(defvar AR 2.8)             ; Asteroid orbit radius
(defvar AS 0.0)             ; Asteroid speed
(defvar TWOPI 6.2831853)
(defvar dt  0.0)            ; Time step, delta-t
(defvar hdt 0.0)            ; Half delta-t

(*proclaim
  '(type (vector-pvar single-float 3) x0 x1 v0 v1 d0 d1 ac wv))

(*defvar x0 (!! #(0.0 0.0 0.0)))    ; positions
(*defvar x1 (!! #(0.0 0.0 0.0)))
(*defvar v0 (!! #(0.0 0.0 0.0)))    ; velocities
(*defvar v1 (!! #(0.0 0.0 0.0)))
(*defvar d0 (!! #(0.0 0.0 0.0)))    ; displacements
(*defvar d1 (!! #(0.0 0.0 0.0)))
(*defvar ac (!! #(0.0 0.0 0.0)))    ; accelerations
(*defvar wv (!! #(0.0 0.0 0.0)))    ; working vector

(*proclaim '(type single-float-pvar m mj r0 r1 w))

(*defvar m  (!! 0.0))    ; masses
(*defvar mj (!! 0.0))    ; working masses
(*defvar r0 (!! 0.0))    ; distances
(*defvar r1 (!! 0.0))
(*defvar w  (!! 0.0))    ; working variable

(*proclaim '(type (unsigned-pvar CN) i j smk))  ; Coordinates, schedule

(*defvar i (!! 0))    (*defvar j (!! 0))
(*defvar smk (!! 0))

(*proclaim '(type (unsigned-pvar *log-number-of-processors-limit*) self))

(*defvar self (self-address!!))  ; working self-address

(*proclaim '(type boolean-pvar act sp lft here smm))

(*defvar act  nil!!)  ; active in accel calculations (off-diag)
(*defvar sp   nil!!)  ; active squares
(*defvar lft  nil!!)  ; left edge
(*defvar here nil!!)  ; active diagonal
(*defvar smm  nil!!)  ; working

;; Global addresses for Paris, also geometry info

(*proclaim
  '(fixnum smk-addr smm-addr self-addr wv-addr sp-addr cube-i cube-i-top))

(defvar smk-addr 0)  (defvar smm-addr 0)    (defvar self-addr 0)
(defvar wv-addr 0)    (defvar sp-addr 0)
(defvar cube-i 0)    (defvar cube-i-top 0)  (defvar cube-i-list nil)

;; Determine cube-address rep for i coordinate

(defun set-cube-i ()
 (setf cube-i (*LISP-I::address-object-cube-address (grid 0 (1- N) 0)))
 (setf cube-i-top (1- (integer-length cube-i)))
 (setf cube-i-list nil)
 (do ((k 0 (incf k)) (kb 1 (setf kb (ash kb 1))))
     ((> k cube-i-top))
   (if (plusp (logand kb cube-i)) (push k cube-i-list)))
)

;; Cold-boot and setup.  Cm need not be cold-booted; uses default if
;; Cm not attached.

(defun setup ()
 (*cold-boot :initial-dimensions '(128 256))
 (setf N 1) (setq LN 0) (do () ((<= NB N)) (setq N (ash N 1)) (incf LN))
 (setf NP (max 1 (truncate cm:*physical-cube-address-limit* (* N N))))
 (setf GEOM (create-geometry
	:dimensions (list NP N N)
	:weights '(0 2 3)
	:ordering '(:news-order :send-order :send-order)))
 (*cold-boot :initial-geometry-definition GEOM)
 (setf smk-addr  (pvar-location smk))  (setf smm-addr (pvar-location smm))
 (setf self-addr (pvar-location self))
 (setf wv-addr   (pvar-location wv))   (setf sp-addr (pvar-location sp))
 (set-cube-i)
 (*set i (self-address-grid!! (!! 1)))
 (*set j (self-address-grid!! (!! 2)))
 (*set smm t!!)  (*set smk (logxor!! i j))
 (do ((k 1 (ash k 1)))
     ((= k N))
   (*when (and!! smm (plusp!! (logand!! (!! k) smk)))
      (*set smk (!! k))  (*set smm nil!!)))
 (*when (=!! (self-address-grid!! (!! 0)) (!! 0))
   (*set sp   t!!)
   (*set act  (and!! (<!! i (!! NB)) (<!! j (!! NB)) (/=!! i j)))
   (*set lft  (zerop!! j))
   (*set here (and!! (=!! i j) (<!! i (!! NB)))))
)

;; Insert asteroids.  This assumes that orbit of radius 1 about
;; origin corresponds to velocity 2*pi (e.g., au-earth-yr units)
;; Uses parallel trigonometry

(defun mkast ()
 (setf NA (- NB NX))
 (when (plusp NA)
  (setf AS (/ TWOPI (sqrt AR)))
  (*when (and!! lft (>=!! i (!! NX)))
    (*set w (/!! (*!! (!! TWOPI) (-!! i (!! NX))) (!! NA)))
    (*let ((sx (sin!! w)) (cx (cos!! w)))
	  (declare (type single-float-pvar sx cx))
	(*setf (aref!! x0 (!! 0)) cx)
	(*setf (aref!! v0 (!! 0)) (-!! sx))
	(*setf (aref!! x0 (!! 1)) sx)
	(*setf (aref!! v0 (!! 1)) cx)
	(*setf (aref!! x0 (!! 2)) (!! 0.0))
	(*setf (aref!! v0 (!! 2)) (!! 0.0)))
    (dsf-vscale!! x0  x0 (!! AR))  (dsf-vscale!! v0  v0 (!! AS))
    (*set m  (!! (/ 1.0 NA)))
  ) ; end *when
 )
)

;; Load initial positions, velocities, also masses, gravitational const.
;; Argument is stream from which to read.

(defun init(strm)
  (dotimes (k NX)
    (dotimes (u 3) (*setf (pref (aref!! x0 (!! u)) (grid 0 k 0)) (read strm))))
  (dotimes (k NX)
    (dotimes (u 3) (*setf (pref (aref!! v0 (!! u)) (grid 0 k 0)) (read strm))))
  (dotimes (k NX) (*setf (pref m (grid 0 k 0)) (read strm)))
  (setf G (read strm))
)

(defun print-state(np)
 (format t
 "~%       x         y         z         vx        vy        vz        m~%")
 (dotimes (k np)
  (let ((gx (grid 0 k 0)))
   (format t "~2D ~9,5F ~9,5F ~9,5F ~9,5F ~9,5F ~9,5F ~9,3E~%"
     k (pref (aref!! x0 (!! 0)) gx)
       (pref (aref!! x0 (!! 1)) gx)
       (pref (aref!! x0 (!! 2)) gx) 
       (pref (aref!! v0 (!! 0)) gx)
       (pref (aref!! v0 (!! 1)) gx)
       (pref (aref!! v0 (!! 2)) gx)
       (pref m gx))))
)

;; Spread vector from diagonal along i axis, leaves result in argument.
;; Uses wv

(*defun spreadi (xv)
  (declare (type (vector-pvar single-float 3) xv))
  (let ((a (+ smk-addr LN)) (b 0)
	(bl cube-i-list) (xv-addr (pvar-location xv)))
       (declare (type fixnum a b))
    (*set smm here)
    (dotimes (k LN)
     (setf b (+ self-addr (pop bl)))
     (CM:move-always CM:context-flag smm-addr 1)
     (CM:logior-always CM:context-flag a 1)
     (decf a)
     (CM:move-constant smm-addr 1 1)
     (CM:lognot b b 1)
     (CM:move wv-addr xv-addr 96);
     (CM:send xv-addr self-addr wv-addr 96)
     (CM:lognot b b 1)
     (CM:move-always CM:context-flag sp-addr 1)
    ) ; end dotimes
   ) ; end let
)


;; Spread from diagonal along j axis, leave result in w
;; Not used at the moment
;;(*defun spreadj (xv)
;;  (declare (type single-float-pvar xv))
;;  (let (a (+ smk-addr LN)) (b (+ self-addr (ash LN -1))))
;;    (*set smm here)
;;    (dotimes (k LN)
;;     (CM:move-always CM:context-flag smm-addr 1)
;;     (CM:logior-always CM:context-flag a 1)
;;     (CM:move-constant smm-addr 1 1)
;;     (CM:lognot b b 1)
;;     (CM:move w-addr xv-addr 32);
;;     (CM:send xv-addr self-addr w-addr 32 nil)
;;     (CM:lognot b b 1)
;;     (CM:move-always CM:context-flag sp-addr 1)
;;    )
;;   (*set w xv)
;;  )
;;)

;; Print NBxNB array for testing

(*defun nbfltpr (x)
 (dotimes (u NB)
   (dotimes (v NB) (format t "~8,2E " (pref (aref!! x (!! 0)) (grid 0 u v))))
   (terpri)
 )
)

;; Calculate initial accelerations

(defun accel1 ()
  (*set ac (!! #(0.0 0.0 0.0)))  (*set x0 (spread!! x0 2 0))
  (*set d0 x0)  (spreadi d0)
  (*when act
    (dsf-v-!! d0 d0 x0) (*set w (sf-vabs-squared!! d0))
    (*set r0 (sqrt!! w))
    (dsf-v/-constant!! ac (sf-v*-constant!! d0 mj) (*!! r0 w))
   ) ; end (*when act ... )
  (*map
   #'(lambda (xc)
      (*locally (declare (type single-float-pvar xc))
       (*set xc (scan!! xc '+!! :dimension 2 :direction :backward)))
      )
   ac)
)

;; Calculate subsequent accelerations

(defun accel ()
  (*set ac (!! #(0.0 0.0 0.0)))  (*set x1 (spread!! x1 2 0))
  (*set d1 x1)  (spreadi d1) 
  (*when act
    (dsf-v-!! d1 d1 x1)  (*set r1 (sf-vabs!! d1))
    (dsf-v/-constant!! ac
     (sf-v*-constant!! (sf-v+!! d1 d0) mj) (*!! r0 r1 (+!! r0 r1)))
   ) ; end (*when act ... )
  (*map
   #'(lambda (xc)
      (*locally (declare (type single-float-pvar xc))
       (*set xc (scan!! xc '+!! :dimension 2 :direction :backward)))
      )
   ac)
)
   

;; Greenspan's 1st-order method.  Performs ns steps of time dt
;; with solution tolerance eps.  Returns number of evaluations.
;; Leaves number of iterations which were stopped at limit in cfail.

(proclaim '(fixnum cfail))
(defvar cfail 0)          ; Counts number of iterations that failed to converge

(defun green1 (ns eps)
 (declare (fixnum ns) (single-float eps))
 (let ((ic 0) (v 0) (lim 10) (mx 0.0))
      (declare (fixnum ic v lim) (single-float mx))
 (*locally (declare (fixnum ns ic v lim) (single-float eps mx))
  (setf hdt (* 0.5 dt))
  (*when sp
    (*set m (spread!! m 2 0))  ;; Spreads scaled masses
    (*if here (*set mj (*!! (!! G) m)) (*set mj (!! -1.0)))
    (*set mj (scan!! mj 'max!! :dimension 1 :direction :forward))
    (*set mj (scan!! mj 'max!! :dimension 1 :direction :backward))
    (dotimes (k ns)
      (setq ic 0)  (accel1)
      (tagbody
	again (incf v) (incf ic)
	      (*when lft
		(dsf-v+!! wv v0 (sf-v*-constant!! ac (!! dt)))
		(setf mx 0.0)
		(*map
		  #'(lambda (df)
		      (*locally (declare (type single-float-pvar df))
			 (setf mx (max mx (*max (abs!! df))))))
		  (sf-v-!! wv v1))
		(*set v1 wv)
		(dsf-v+!! wv x0 (sf-v*-constant!! (sf-v+!! v0 v1) (!! hdt)))
		(*map
		  #'(lambda (df)
		      (*locally (declare (type single-float-pvar df))
			 (setf mx (max mx (*max (abs!! df))))))
		  (sf-v-!! wv x1))
		(*set x1 wv))
	 (if (or (< mx eps) (= ic lim)) (go finished))
	      (accel) (go again)
         finished
       ) ; end tagbody
      (*when lft (*set x0 x1) (*set v0 v1))
      (if (= ic lim) (incf cfail))
    ) ; end dotimes
   )  ; end *locally
   ) v) ; end let
)

(defun ax-pstate(tm)
 (declare (single-float tm))
 (let ((np (cond ((zerop PX) 0) ((= PX 1) NX) (NB))))
    (if (and (< PX 3) (> np 0))
      (print-state np)
      (when (>= PX 3) (format t "No plotting yet~%")))
 )
)

;; Run integration sequence ns time steps per cycle, np cycles of output
;; time step dt, tolerance eps

(defun rung (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
 (let ((simtm 0.0) (vl 0))
  (setf dt xdt)
  (format t "~%   G = ~13,6E  delta-t = ~12,5F  epsilon = ~12,5E~%"
		  G dt eps)
  (format t "   ~D asteroids at radius = ~9,5F~%" NA AR)
  (format t "~% time = 0")  (ax-pstate 0.0)
  (dotimes (k np)
    (setf cfail 0)
    (setf simtm (* (1+ k) ns dt))
    (if (<= PX 2)
      (progn (cm:time (setf vl (green1 ns eps)))
	     (format t
         "~% simulated time = ~9,3F  ~D evaluations  ~D iterations limited~%"
			simtm vl cfail))
      (setf vl (green1 ns eps)))
    (ax-pstate simtm)
   ) ; end dotimes
 )
)

(defvar SX 0.0)        ; Scale values for later plotting
(defvar SY 0.0)

;; Manage calculation

(defun main ()
  (let (ns np cmd nv dt eps file strm)
    (setf PX 2)
    (tagbody
      again (format t "nbax> ") (setf cmd (read))
	(case cmd
	  (q (go finished))
	  (p (setf PX (read)))
	  (s (setf SX (read)) (setf SY (read)))
	  (n (setq NB (read)))
	  (c (rung ns np dt eps))
	  (r (go run))
	  (T (format t "Unrecognized command~%")))
	(go again)
     run
	(format t "Data file> ")  (setf file (read-line))
	(setf strm (open file :direction :input))
	(setup)
	(setf NX (read strm))  (setf AR (read strm))  (init strm)
	(mkast)
	(setf ns (read strm))  (setf np  (read strm))
	(setf dt (read strm))  (setf eps (read strm))
	(when (>= PX 3) (format t "No plotting yet~%"))
	(close strm)
	(rung ns np dt eps)
	(go again)
     finished)
  )
)
